@@ -15,75 +15,6 @@ import sys
 
 from tools import shared
 
-if len(sys.argv) < 2 or sys.argv[1] in ['-v', '-help', '--help', '-?', '?']:
-  print('''
-Emscripten System Builder Tool
-==============================
-
-You can use this tool to manually build parts of the emscripten system
-environment. In general emcc will build them automatically on demand, so
-you do not strictly need to use this tool, but it gives you more control
-over the process (in particular, if emcc does this automatically, and you
-are running multiple build commands in parallel, confusion can occur).
-
-Usage:
-
-  embuilder.py OPERATION TASK1 [TASK2..]
-
-Available operations and tasks:
-
-  build libc
-        libc-mt
-        libc-extras
-        struct_info
-        emmalloc
-        emmalloc_debug
-        dlmalloc
-        dlmalloc_debug
-        dlmalloc_threadsafe
-        dlmalloc_threadsafe_debug
-        dlmalloc_noerrno
-        dlmalloc_debug_noerrno
-        dlmalloc_threadsafe_noerrno
-        dlmalloc_threadsafe_debug_noerrno
-        pthreads
-        libc++
-        libc++_noexcept
-        libc++abi
-        gl
-        gl-mt
-        native_optimizer
-        binaryen
-        bullet
-        freetype
-        libpng
-        ogg
-        sdl2
-        sdl2-image
-        sdl2-mixer
-        sdl2-ttf
-        sdl2-net
-        vorbis
-        zlib
-        cocos2d
-        libc-wasm
-        compiler_rt_wasm
-        regal
-
-Issuing 'embuilder.py build ALL' causes each task to be built.
-
-It is also possible to build native_optimizer manually by using CMake. To
-do that, run
-
-   1. cd $EMSCRIPTEN/tools/optimizer
-   2. cmake . -DCMAKE_BUILD_TYPE=Release
-   3. make (or mingw32-make/vcbuild/msbuild on Windows)
-
-and set up the location to the native optimizer in ~/.emscripten
-
-''')
-  sys.exit(0)
-
 C_BARE = '''
         int main() {}
       '''
@@ -109,24 +40,100 @@ CXX_WITH_STDLIB = '''
       '''
 
 SYSTEM_TASKS = [
-    'al', 'compiler-rt', 'gl', 'gl-mt', 'libc', 'libc-mt', 'libc-extras',
-    'emmalloc', 'emmalloc_debug', 'dlmalloc', 'dlmalloc_threadsafe', 'pthreads',
-    'dlmalloc_debug', 'dlmalloc_threadsafe_debug', 'libc++', 'libc++_noexcept',
-    'dlmalloc_debug_noerrno', 'dlmalloc_threadsafe_debug_noerrno',
-    'dlmalloc_noerrno', 'dlmalloc_threadsafe_noerrno',
-    'libc++abi', 'html5'
+    'al',
+    'compiler-rt',
+    'emmalloc',
+    'emmalloc_debug',
+    'gl',
+    'gl-emu',
+    'gl-emu-webgl2',
+    'gl-mt',
+    'gl-mt-emu',
+    'gl-mt-emu-webgl2',
+    'gl-mt-webgl2',
+    'gl-webgl2',
+    'html5',
+    'libc',
+    'libc++',
+    'libc++_noexcept',
+    'libc++abi',
+    'libc-extras',
+    'libc-mt',
+    'pthreads',
+    'pthreads_stub',
 ]
+
+for debug in ['', '_debug']:
+  for noerrno in ['', '_noerrno']:
+    for threadsafe in ['', '_threadsafe']:
+      for tracing in ['', '_tracing']:
+        SYSTEM_TASKS += ['dlmalloc' + debug + noerrno + threadsafe + tracing]
+
 USER_TASKS = [
-    'binaryen', 'bullet', 'freetype', 'icu', 'libpng', 'ogg', 'sdl2',
-    'sdl2-gfx', 'sdl2-image', 'sdl2-mixer', 'sdl2-ttf', 'sdl2-net',
-    'vorbis', 'zlib', 'regal'
+    'binaryen',
+    'bullet',
+    'cocos2d',
+    'freetype',
+    'icu',
+    'libpng',
+    'ogg',
+    'regal',
+    'sdl2',
+    'sdl2-gfx',
+    'sdl2-image',
+    'sdl2-image-png',
+    'sdl2-mixer',
+    'sdl2-net',
+    'sdl2-ttf',
+    'vorbis',
+    'zlib',
 ]
 
 temp_files = shared.configuration.get_temp_files()
 logger = logging.getLogger('embuilder')
 
 
+def print_help():
+  all_tasks = SYSTEM_TASKS + USER_TASKS
+  all_tasks.sort()
+  print('''
+Emscripten System Builder Tool
+==============================
+
+You can use this tool to manually build parts of the emscripten system
+environment. In general emcc will build them automatically on demand, so
+you do not strictly need to use this tool, but it gives you more control
+over the process (in particular, if emcc does this automatically, and you
+are running multiple build commands in parallel, confusion can occur).
+
+Usage:
+
+  embuilder.py OPERATION TASK1 [TASK2..] [--lto]
+
+Available operations and tasks:
+
+  build %s
+
+Issuing 'embuilder.py build ALL' causes each task to be built.
+
+Flags:
+
+  --lto  Build bitcode files, for LTO with the LLVM wasm backend
+
+It is also possible to build native_optimizer manually by using CMake. To
+do that, run
+
+   1. cd $EMSCRIPTEN/tools/optimizer
+   2. cmake . -DCMAKE_BUILD_TYPE=Release
+   3. make (or mingw32-make/vcbuild/msbuild on Windows)
+
+and set up the location to the native optimizer in ~/.emscripten
+''' % '\n        '.join(all_tasks))
+
+
 def build(src, result_libs, args=[]):
+  if shared.Settings.WASM_BACKEND:
+    args = args + ['-s', 'WASM_OBJECT_FILES=%d' % shared.Settings.WASM_OBJECT_FILES]
   # build in order to generate the libraries
   # do it all in a temp dir where everything will be cleaned up
   temp_dir = temp_files.get_dir()
@@ -148,13 +155,43 @@ def build_port(port_name, lib_name, params):
   build(C_BARE, [os.path.join('ports-builds', port_name, lib_name)] if lib_name else [], params)
 
 
+def static_library_name(name):
+  if shared.Settings.WASM_BACKEND and shared.Settings.WASM_OBJECT_FILES:
+    return name + '.a'
+  else:
+    return name + '.bc'
+
+
 def main():
+  if len(sys.argv) < 2 or sys.argv[1] in ['-v', '-help', '--help', '-?', '?']:
+    print_help()
+    return 0
+
   operation = sys.argv[1]
   if operation != 'build':
     shared.exit_with_error('unfamiliar operation: ' + operation)
 
+  # process flags
+
+  args = sys.argv[2:]
+
+  def is_flag(arg):
+    return arg.startswith('--')
+
+  for arg in args:
+    if is_flag(arg):
+      arg = arg[2:]
+      if arg == 'lto':
+        shared.Settings.WASM_OBJECT_FILES = 0
+        # Reconfigure the cache dir to reflect the change
+        shared.reconfigure_cache()
+
+  args = [a for a in args if not is_flag(a)]
+
+  # process tasks
+
   auto_tasks = False
-  tasks = sys.argv[2:]
+  tasks = args
   if 'SYSTEM' in tasks:
     tasks = SYSTEM_TASKS
     auto_tasks = True
@@ -163,7 +200,10 @@ def main():
     auto_tasks = True
   if auto_tasks:
     if shared.Settings.WASM_BACKEND:
-      skip_tasks = {'libc-mt', 'gl-mt', 'dlmalloc_threadsafe', 'dlmalloc_threadsafe_debug', 'pthreads'}
+      skip_tasks = [task for task in SYSTEM_TASKS + USER_TASKS if '-mt' in task or 'thread' in task]
+      # cocos2d: must be ported, errors on
+      # "Cannot recognize the target platform; are you targeting an unsupported platform?"
+      skip_tasks += ['cocos2d']
       print('Skipping building of %s, because WebAssembly does not support pthreads.' % ', '.join(skip_tasks))
       tasks = [x for x in tasks if x not in skip_tasks]
     else:
@@ -185,38 +225,35 @@ def main():
         }
       ''', ['libcompiler_rt.a'])
     elif what == 'libc':
-      build(C_WITH_MALLOC, ['libc.bc'])
+      build(C_WITH_MALLOC, [static_library_name('libc')])
     elif what == 'libc-extras':
       build('''
         extern char **environ;
         int main() {
           return (int)environ;
         }
-      ''', ['libc-extras.bc'])
+      ''', [static_library_name('libc-extras')])
     elif what == 'struct_info':
       build(C_BARE, ['generated_struct_info.json'])
     elif what == 'emmalloc':
-      build(C_WITH_MALLOC, ['libemmalloc.bc'], ['-s', 'MALLOC="emmalloc"'])
+      build(C_WITH_MALLOC, [static_library_name('libemmalloc')], ['-s', 'MALLOC="emmalloc"'])
     elif what == 'emmalloc_debug':
-      build(C_WITH_MALLOC, ['libemmalloc_debug.bc'], ['-s', 'MALLOC="emmalloc"', '-g'])
-    elif what == 'dlmalloc':
-      build(C_WITH_MALLOC, ['libdlmalloc.bc'], ['-s', 'MALLOC="dlmalloc"'])
-    elif what == 'dlmalloc_debug':
-      build(C_WITH_MALLOC, ['libdlmalloc_debug.bc'], ['-g', '-s', 'MALLOC="dlmalloc"'])
-    elif what == 'dlmalloc_threadsafe_debug':
-      build(C_WITH_MALLOC, ['libdlmalloc_threadsafe_debug.bc'], ['-g', '-s', 'USE_PTHREADS=1', '-s', 'MALLOC="dlmalloc"'])
-    elif what in ('dlmalloc_threadsafe', 'libc-mt', 'pthreads'):
-      build(C_WITH_MALLOC, ['libc-mt.bc', 'libdlmalloc_threadsafe.bc', 'libpthreads.bc'], ['-s', 'USE_PTHREADS=1', '-s', 'MALLOC="dlmalloc"'])
-    elif what == 'dlmalloc_noerrno':
-      build(C_WITH_MALLOC, ['libdlmalloc_noerrno.bc'], ['-s', 'MALLOC="dlmalloc"', '-s', 'SUPPORT_ERRNO=0'])
-    elif what == 'dlmalloc_debug_noerrno':
-      build(C_WITH_MALLOC, ['libdlmalloc_debug_noerrno.bc'], ['-g', '-s', 'MALLOC="dlmalloc"', '-s', 'SUPPORT_ERRNO=0'])
-    elif what == 'dlmalloc_threadsafe_debug_noerrno':
-      build(C_WITH_MALLOC, ['libdlmalloc_threadsafe_debug_noerrno.bc'], ['-g', '-s', 'USE_PTHREADS=1', '-s', 'MALLOC="dlmalloc"', '-s', 'SUPPORT_ERRNO=0'])
-    elif what == 'dlmalloc_threadsafe_noerrno':
-      build(C_WITH_MALLOC, ['libdlmalloc_threadsafe_noerrno.bc'], ['-s', 'USE_PTHREADS=1', '-s', 'MALLOC="dlmalloc"', '-s', 'SUPPORT_ERRNO=0'])
+      build(C_WITH_MALLOC, [static_library_name('libemmalloc_debug')], ['-s', 'MALLOC="emmalloc"', '-g'])
+    elif what.startswith('dlmalloc'):
+      cmd = ['-s', 'MALLOC="dlmalloc"']
+      if '_debug' in what:
+        cmd += ['-g']
+      if '_noerrno' in what:
+        cmd += ['-s', 'SUPPORT_ERRNO=0']
+      if '_threadsafe' in what:
+        cmd += ['-s', 'USE_PTHREADS=1']
+      if '_tracing' in what:
+        cmd += ['-s', 'EMSCRIPTEN_TRACING=1']
+      build(C_WITH_MALLOC, [static_library_name('lib' + what)], cmd)
+    elif what in ('libc-mt', 'pthreads'):
+      build(C_WITH_MALLOC, [static_library_name('libc-mt'), static_library_name('libpthreads')], ['-s', 'USE_PTHREADS=1'])
     elif what == 'libc-wasm':
-      build(C_WITH_STDLIB, ['libc-wasm.bc'], ['-s', 'WASM=1'])
+      build(C_WITH_STDLIB, [static_library_name('libc-wasm')], ['-s', 'WASM=1'])
     elif what == 'libc++':
       build(CXX_WITH_STDLIB, ['libc++.a'], ['-s', 'DISABLE_EXCEPTION_CATCHING=0'])
     elif what == 'libc++_noexcept':
@@ -230,21 +267,21 @@ def main():
           y->a();
           return y->y;
         }
-      ''', ['libc++abi.bc'])
-    elif what == 'gl':
+      ''', [static_library_name('libc++abi')])
+    elif what == 'gl' or what.startswith('gl-'):
+      opts = []
+      if '-mt' in what:
+        opts += ['-s', 'USE_PTHREADS=1']
+      if '-emu' in what:
+        opts += ['-s', 'LEGACY_GL_EMULATION=1']
+      if '-webgl2' in what:
+        opts += ['-s', 'USE_WEBGL2=1']
       build('''
         extern "C" { extern void* emscripten_GetProcAddress(const char *x); }
         int main() {
           return int(emscripten_GetProcAddress("waka waka"));
         }
-      ''', ['libgl.bc'])
-    elif what == 'gl-mt':
-      build('''
-        extern "C" { extern void* emscripten_GetProcAddress(const char *x); }
-        int main() {
-          return int(emscripten_GetProcAddress("waka waka"));
-        }
-      ''', ['libgl-mt.bc'], ['-s', 'USE_PTHREADS=1'])
+      ''', [static_library_name('lib' + what)], opts)
     elif what == 'native_optimizer':
       build(C_BARE, ['optimizer.2.exe'], ['-O2', '-s', 'WASM=0'])
     elif what == 'compiler_rt_wasm':
@@ -260,7 +297,15 @@ def main():
           return emscripten_compute_dom_pk_code(NULL);
         }
 
-      ''', ['libhtml5.bc'])
+      ''', [static_library_name('libhtml5')])
+    elif what == 'pthreads_stub':
+      build('''
+        #include <emscripten/threading.h>
+        int main() {
+          return emscripten_is_main_runtime_thread();
+        }
+
+      ''', [static_library_name('libpthreads_stub')])
     elif what == 'al':
       build('''
         #include "AL/al.h"
@@ -268,7 +313,7 @@ def main():
           alGetProcAddress(0);
           return 0;
         }
-      ''', ['libal.bc'])
+      ''', [static_library_name('libal')])
     elif what == 'icu':
       build_port('icu', 'libicuuc.bc', ['-s', 'USE_ICU=1'])
     elif what == 'zlib':
@@ -287,6 +332,8 @@ def main():
       build_port('sdl2-gfx', 'libsdl2_gfx.bc', ['-s', 'USE_SDL=2', '-s', 'USE_SDL_IMAGE=2', '-s', 'USE_SDL_GFX=2'])
     elif what == 'sdl2-image':
       build_port('sdl2-image', 'libsdl2_image.bc', ['-s', 'USE_SDL=2', '-s', 'USE_SDL_IMAGE=2'])
+    elif what == 'sdl2-image-png':
+      build_port('sdl2-image', 'libsdl2_image.bc', ['-s', 'USE_SDL=2', '-s', 'USE_SDL_IMAGE=2', '-s', 'SDL2_IMAGE_FORMATS=["png"]'])
     elif what == 'sdl2-net':
       build_port('sdl2-net', 'libsdl2_net.bc', ['-s', 'USE_SDL=2', '-s', 'USE_SDL_NET=2'])
     elif what == 'sdl2-mixer':
@@ -305,7 +352,7 @@ def main():
       build_port('regal', 'libregal.bc', ['-s', 'USE_REGAL=1'])
     else:
       logger.error('unfamiliar build target: ' + what)
-      sys.exit(1)
+      return 1
 
     logger.info('...success')
   return 0
